import json
import os
from argparse import ArgumentParser

import boto3
import dateutil.parser
import pandas as pd
from google.cloud import storage

from ugbio_omics.cromwell_calculate_cost import calculate_cost
from ugbio_omics.get_preformance_and_cost import performance as omics_performance
from ugbio_omics.get_run_cost import Columns, RunCost
from ugbio_omics.get_run_info import get_run_info


def compare_cromwell_omics(cromwell_wid, omics_run_id, omics_session, workflow_name, output_path, *, overwrite=False):
    # Create the output directory if it doesn't exist
    if output_path:
        os.makedirs(output_path, exist_ok=True)
    else:
        output_path = os.getcwd()

    # Get cromwell performance and metadata files
    cromwell_performance_file, cromwell_metadata_file = copy_cromwell_data(
        workflow_name, cromwell_wid, output_path, overwrite=overwrite
    )
    # Calculate cromwell cost
    cromwell_cost_df, cromwell_disk_cost = cromwell_cost(
        cromwell_metadata_file, output_path, workflow_name, cromwell_wid
    )
    # Calculate cromwell performance
    cromwell_performance_df = cromwell_performance(cromwell_performance_file)

    # Calculate omics performance and cost
    omics_performance_df, omics_disk_cost, omics_run_cost = get_omics_performance_cost(
        omics_run_id, omics_session, output_path
    )

    # Compare cromwell and omics cost
    print("Comparing cromwell and omics...")
    cost_df = pd.DataFrame(omics_performance_df, columns=["task", "cost_SUM"])
    cost_df = cost_df.merge(cromwell_cost_df[["task", "compute_cost"]], on="task", how="outer")
    cost_df = cost_df.rename(columns={"cost_SUM": "omics_cost", "compute_cost": "cromwell_cost"})
    # add disk cost
    cost_df = pd.concat(
        [
            cost_df,
            pd.DataFrame({"task": ["disk"], "cromwell_cost": [cromwell_disk_cost], "omics_cost": [omics_disk_cost]}),
        ],
        ignore_index=True,
    )
    # add cost difference
    cost_df["cost_diff"] = cost_df["omics_cost"] - cost_df["cromwell_cost"]
    # move total cost to the end
    cost_df = pd.concat([cost_df[cost_df.task != "total"], cost_df[cost_df.task == "total"]], ignore_index=True)

    # Compare cromwell and omics run duration
    cromwell_total_duration = get_cromwell_total_duration(cromwell_metadata_file)
    cromwell_performance_df = pd.concat(
        [cromwell_performance_df, pd.DataFrame({"task": ["total"], "run_time (hours)": [cromwell_total_duration]})],
        ignore_index=True,
    )
    omics_total_duration = get_omics_total_duration(omics_run_id, omics_session)
    omics_performance_df.loc[omics_performance_df["task"] == "total", "run_time (hours)"] = omics_total_duration
    duration_df = pd.DataFrame(omics_performance_df, columns=["task", "run_time (hours)"])
    duration_df = duration_df.merge(cromwell_performance_df[["task", "run_time (hours)"]], on="task", how="outer")
    duration_df = duration_df.rename(
        columns={"run_time (hours)_x": "omics_duration", "run_time (hours)_y": "cromwell_duration"}
    )
    # add duration difference
    duration_df["duration_diff"] = duration_df["omics_duration"] - duration_df["cromwell_duration"]

    # Add resources for omics and cromwell
    resources_df = extract_omics_resources(omics_run_cost)

    # Merge cromwell_cost_df['cpus','mem_gbs'] into one column as a dictionary
    cromwell_resources = cromwell_cost_df[["cpus", "mem_gbs"]].apply(
        lambda x: {"cpus": x["cpus"], "mem_gbs": x["mem_gbs"]}, axis=1
    )
    cromwell_cost_df["cromwell_resources"] = cromwell_resources
    resources_df = resources_df.merge(cromwell_cost_df[["task", "cromwell_resources"]], on="task", how="outer")

    # Merge cost and duration and save to file
    compare_df = cost_df.merge(duration_df, on="task", how="outer")
    compare_df = compare_df.merge(resources_df, on="task", how="outer")
    compare_file = f"{output_path}/compare_omics_{omics_run_id}_cromwell_{cromwell_wid}.csv"
    compare_df.to_csv(compare_file, index=False, float_format="%.3f")
    print(f"Comparison saved in: {compare_file}")


def extract_omics_resources(omics_run_cost: RunCost) -> pd.DataFrame:
    resources_df = omics_run_cost.get_tasks_resources()
    # Merge cpus, mem_gbs, gpus into one column as a dictionary
    resources_df["omics_resources"] = resources_df.apply(
        lambda row: {
            "cpus": row[Columns.CPU_REQUESTED.value],
            "mem_gbs": row[Columns.MEMORY_REQUESTED_GIB.value],
            "gpus": row[Columns.GPUS_REQUESTED.value],
        },
        axis=1,
    )
    resources_df = resources_df.drop(
        columns=[Columns.CPU_REQUESTED.value, Columns.MEMORY_REQUESTED_GIB.value, Columns.GPUS_REQUESTED.value],
    )
    resources_df = resources_df.rename(
        columns={Columns.NAME_COLUMN.value: "task", Columns.OMICS_INSTANCE_TYPE_RESERVED.value: "omics_instance"},
    )
    resources_df["task"] = resources_df["task"].str.split("-").str[0]
    resources_df = resources_df.groupby("task").first().reset_index()
    return resources_df


def copy_cromwell_data(workflow_name, cromwell_wid, output_path, *, overwrite=False):
    # Get cromwell bucket for the workflow
    print(f"Get cromwell performance file for workflow: {workflow_name}/{cromwell_wid}")
    cromwell_client = storage.Client()
    bucket_name = "cromwell-backend-ultima-data-307918"
    bucket = cromwell_client.get_bucket(bucket_name)

    # save performance.csv as local file
    performance_file = f"{output_path}/cromwell_{cromwell_wid}.performance.csv"
    if os.path.exists(performance_file) and not overwrite:
        print(f"Skipping download. Cromwell performance file already exists: {performance_file}")
    else:
        blob = bucket.get_blob(f"cromwell-execution/{workflow_name}/{cromwell_wid}/performance.csv")
        if not blob:
            raise Exception(f"Performance file not found for workflow: {workflow_name}/{cromwell_wid}")
        with open(performance_file, "w") as f:
            f.write(blob.download_as_text(encoding="utf-8"))
        print(f"Cromwell performance saved in: {performance_file}")

    # save metadata.json as local file
    print(f"Get cromwell metadata file for workflow: {workflow_name}/{cromwell_wid}")
    metadata_file = f"{output_path}/cromwell_{cromwell_wid}.metadata.json"
    if os.path.exists(metadata_file) and not overwrite:
        print(f"Skipping download. Cromwell metadata file already exists: {metadata_file}")
    else:
        blob = bucket.get_blob(f"cromwell-execution/{workflow_name}/{cromwell_wid}/metadata.json")
        if not blob:
            raise Exception(f"Metadata file not found for workflow: {workflow_name}/{cromwell_wid}")
        with open(metadata_file, "w") as f:
            f.write(blob.download_as_text(encoding="utf-8"))
        print(f"Cromwell metadata saved in: {metadata_file}")

    return performance_file, metadata_file


def cromwell_cost(metadata_file, output_path, workflow_name, cromwell_wid):
    # run the calculate_cost.py script from terra_pipeline
    print(f"Calculating cromwell cost for workflow: {workflow_name}/{cromwell_wid}")

    cromwell_cost_file = f"{output_path}/cromwell_{cromwell_wid}.cost.csv"

    with open(metadata_file) as data_file:
        metadata = json.load(data_file)

    try:
        calculate_cost(
            metadata=metadata,
            ignore_preempted=False,
            only_total_cost=False,
            print_header=True,
            output_file=cromwell_cost_file,
        )
    except Exception as e:
        print(f"Error occurred while calculating cromwell cost: {e}")
        raise e

    print(f"Cromwell cost saved in: {cromwell_cost_file}")

    # load cromwell cost into a dataframe and clean up the data
    cromwell_cost_df = pd.read_csv(cromwell_cost_file)
    cromwell_cost_df = cromwell_cost_df.rename(columns=lambda x: x.strip())
    # if there is a any row similiar to the column names- delete that row from df and re-save the file
    cromwell_cost_df = cromwell_cost_df[~cromwell_cost_df["task_name"].str.contains("task_name")]
    # convert all numeric columns to numeric
    for col in cromwell_cost_df.columns:
        if "cost" in col:
            cromwell_cost_df[col] = pd.to_numeric(cromwell_cost_df[col], errors="coerce")
    # numeric_columns = ["cpu_cost", "pe_cpu_cost", "mem_cost", "pe_mem_cost", "gpu_cost", "pe_gpu_cost", "disk_cost"]
    # cromwell_cost_df[numeric_columns] = cromwell_cost_df[numeric_columns].apply(pd.to_numeric, errors='coerce')

    # add compute cost as an additional column
    cromwell_cost_df["compute_cost"] = (
        cromwell_cost_df["cpu_cost"]
        + cromwell_cost_df["pe_cpu_cost"]
        + cromwell_cost_df["mem_cost"]
        + cromwell_cost_df["pe_mem_cost"]
        + cromwell_cost_df["gpu_cost"]
        + cromwell_cost_df["pe_gpu_cost"]
    )
    cromwell_cost_df["task"] = cromwell_cost_df["task_name"].str.split(".").str[1]

    # add total cost as an additional row
    cromwell_total_cost = cromwell_cost_df["total_cost"].sum()
    cromwell_cost_df = pd.concat(
        [cromwell_cost_df, pd.DataFrame({"task": ["total"], "compute_cost": [cromwell_total_cost]})], ignore_index=True
    )

    cromwell_disk_cost = cromwell_cost_df["disk_cost"].sum()

    return cromwell_cost_df, cromwell_disk_cost


def cromwell_performance(cromwell_performance_file):
    # load and clean cromwell performance into a dataframe
    performance_df = pd.read_csv(cromwell_performance_file)
    performance_df = performance_df.rename(columns=lambda x: x.strip())
    performance_df["task"] = performance_df["task"].str.strip()
    numeric_columns = performance_df.columns.drop(["task"])
    performance_df[numeric_columns] = performance_df[numeric_columns].apply(pd.to_numeric, errors="coerce")

    # Drop multiple preemtible attempts and keep the last one
    performance_df["attempt"] = performance_df["task"].str.extract(r"attempt-(\d+)", expand=False).astype(float)
    performance_df["task"] = performance_df["task"].str.split("/attempt-", expand=True)[0]
    performance_df = performance_df.sort_values("attempt", ascending=False).drop_duplicates(subset="task", keep="first")
    performance_df = performance_df.drop(columns="attempt")

    # group by task name and calculate mean of performance metrics
    performance_df["task"] = performance_df["task"].str.split("/").str[0]
    grouped_df = performance_df.groupby("task").mean().reset_index()

    # remove prefix "call-" from task name
    grouped_df["task"] = grouped_df["task"].str.split("call-").str[1]

    return grouped_df


def get_cromwell_total_duration(cromwell_metadata_file):
    with open(cromwell_metadata_file) as f:
        metadata = json.load(f)
    end_time = dateutil.parser.parse(metadata["end"])
    start_time = dateutil.parser.parse(metadata["start"])
    total_duration = end_time - start_time
    return total_duration.total_seconds() / 3600


def get_omics_performance_cost(
    omics_run_id, session, output_path, *, overwrite=False
) -> tuple[pd.DataFrame, float, RunCost]:
    print(f"Calculating omics performance and cost for run: {omics_run_id}")
    performance_file = f"{output_path}/omics_{omics_run_id}.performance.csv"
    if os.path.exists(performance_file) and not overwrite:
        print(f"Skipping download. Omics performance file already exists: {performance_file}")
        performance_df = pd.read_csv(performance_file)
        run_cost = RunCost(omics_run_id, output_dir=output_path)
    else:
        performance_df, run_cost = omics_performance(omics_run_id, session=session, output_dir=output_path)
    performance_df = performance_df.rename(columns=lambda x: x.strip())
    performance_df["task"] = performance_df["task"].str.strip()

    omics_disk_cost = run_cost.get_storage_cost()

    # group by task name, calculate mean of performance metrics and sum of cost
    performance_df["task"] = performance_df["task"].str.split("-").str[0]
    # Convert all columns except 'task' and 'instance' to numeric
    for col in performance_df.columns:
        if col not in ["task", "instance"]:
            performance_df[col] = pd.to_numeric(performance_df[col], errors="coerce")
    # group by task name and calculate mean of performance metrics
    grouped_df = performance_df.groupby("task").mean(numeric_only=True).drop("cost", axis=1)
    grouped_df["cost_SUM"] = performance_df.groupby("task")["cost"].sum()
    grouped_df = grouped_df.reset_index()
    # add total cost as an additional row
    total_run_cost = run_cost.get_total_cost()
    grouped_df = pd.concat(
        [grouped_df, pd.DataFrame({"task": ["total"], "cost_SUM": [total_run_cost]})], ignore_index=True
    )

    return grouped_df, omics_disk_cost, run_cost


def get_omics_total_duration(omics_run_id, session=None):
    if session:
        omics_client = session.client("omics")
    else:
        omics_client = boto3.client("omics")

    run = get_run_info(omics_run_id, client=omics_client)
    total_duration = run["duration"].total_seconds() / 3600
    return total_duration


def main():
    parser = ArgumentParser()
    # cromwell args
    parser.add_argument("--cromwell-wid", type=str, help="Cromwell workflow id to analyze")

    # omics args
    parser.add_argument("--region", type=str, help="AWS region to use", default="us-east-1")
    parser.add_argument("--omics-run-id", type=str, help="HealthOmics workflow run-id to analyze")

    # general args
    parser.add_argument("--workflow-name", type=str, help="Workflow name e.g. EfficientDV")
    parser.add_argument(
        "--output-path",
        type=str,
        help="Output path for all files copied and generated during the analysis",
        default=None,
    )
    parser.add_argument("--overwrite", type=bool, help="Overwrite downloaded files from cloud", default=False)

    args = parser.parse_args()
    session = boto3.Session(region_name=args.region)

    compare_cromwell_omics(
        args.cromwell_wid, args.omics_run_id, session, args.workflow_name, args.output_path, overwrite=args.overwrite
    )


if __name__ == "__main__":
    main()

# output_path = "/home/inbalzelig/data/omics/somatic_efficientdv"
# workflow_name = "EfficientDV"
# cromwell_wid = "31b29c63-f7b2-4a7c-986a-64171126d9c4"
# omics_run_id = "2068853"



from argparse import ArgumentParser

import boto3
import pandas as pd

from ugbio_omics.get_omics_log import OMICS_LOG_GROUP, fetch_save_log


def parse_manifest_log(run_id, output_path=None, session=None, output_prefix=""):
    manifest_json_file = f"{output_prefix}omics_{run_id}_manifest.json"

    # Get logs client
    if session:
        client = session.client("logs")
    else:
        client = boto3.client("logs")

    # get manifest log stream name
    res = client.describe_log_streams(logGroupName=OMICS_LOG_GROUP, logStreamNamePrefix=f"manifest/run/{run_id}")

    if not res.get("logStreams"):
        print(f"No manifest log stream found for run id '{run_id}'")
        return

    log_stream_name = res["logStreams"][0]["logStreamName"]

    manifest_json_file = fetch_save_log(log_stream_name, manifest_json_file, output_path, session)

    # parse manifest log from json into df
    manifest_df = pd.read_json(manifest_json_file, lines=True)

    # Save general run info and print out storage info
    run_df = manifest_df.head(1).dropna(axis=1, how="all")

    print("---------Storage info---------")
    storage_df = run_df["metrics"].apply(pd.Series)
    print(storage_df.to_json(orient="records", lines=True, default_handler=str, indent=4))

    general_run_info_file = f"{output_prefix}omics_{run_id}_general_run_info.json"
    if output_path:
        general_run_info_file = f"{output_path}/{general_run_info_file}"
    print(f"Saving general run info to: {general_run_info_file}")
    run_df.to_json(general_run_info_file, orient="records", lines=True, default_handler=str, indent=4)

    # Save tasks info to csv
    tasks_df = manifest_df.drop(0).dropna(axis=1, how="all")  # remove first line that contains general run info
    metrics_df = tasks_df["metrics"].apply(pd.Series)  # expand metrics column into separate columns
    tasks_df = pd.concat([tasks_df.drop(["metrics"], axis=1), metrics_df], axis=1)

    tasks_manifest_csv_file = f"{output_prefix}omics_{run_id}_task_manifests.csv"
    if output_path:
        tasks_manifest_csv_file = f"{output_path}/{tasks_manifest_csv_file}"
    print(f"Saving task manifests to: {tasks_manifest_csv_file}")
    tasks_df.to_csv(tasks_manifest_csv_file, index=False)


def main():
    parser = ArgumentParser()
    parser.add_argument("--region", type=str, help="AWS region to use", default="us-east-1")
    parser.add_argument("--run-id", type=str, help="HealthOmics workflow run-id to analyze")
    parser.add_argument("--output", type=str, help="Output dir to save log events", default=None)
    parser.add_argument("--output-prefix", type=str, help="File name prefix for the output", required=False, default="")

    args = parser.parse_args()
    session = boto3.Session(region_name=args.region)

    parse_manifest_log(args.run_id, output_path=args.output, session=session, output_prefix=args.output_prefix)


if __name__ == "__main__":
    main()



import os
from argparse import ArgumentParser

import boto3
import pandas as pd

from ugbio_omics.compare_cromwell_omics import (
    extract_omics_resources,
    get_omics_performance_cost,
    get_omics_total_duration,
)


def single_run(omics_run_id, omics_session, output_path):
    # Calculate omics performance and cost
    omics_performance_df, omics_disk_cost, omics_run_cost = get_omics_performance_cost(
        omics_run_id, omics_session, output_path
    )

    # fix up cost data
    cost_df = pd.DataFrame(omics_performance_df, columns=["task", "cost_SUM"])
    cost_df = cost_df.rename(columns={"cost_SUM": f"{omics_run_id}_cost"})
    # add disk cost
    cost_df = pd.concat(
        [cost_df, pd.DataFrame({"task": ["disk"], f"{omics_run_id}_cost": [omics_disk_cost]})], ignore_index=True
    )
    # move total cost to the end
    cost_df = pd.concat([cost_df[cost_df.task != "total"], cost_df[cost_df.task == "total"]], ignore_index=True)

    # run duration
    omics_total_duration = get_omics_total_duration(omics_run_id, omics_session)
    omics_performance_df.loc[omics_performance_df["task"] == "total", "run_time (hours)"] = omics_total_duration
    duration_df = pd.DataFrame(omics_performance_df, columns=["task", "run_time (hours)"])
    duration_df = duration_df.rename(columns={"run_time (hours)": f"{omics_run_id}_duration"})

    # Add resources
    resources_df = extract_omics_resources(omics_run_cost)
    resources_df = resources_df.rename(
        columns={"omics_resources": f"{omics_run_id}_resources", "omics_instance": f"{omics_run_id}_instance"}
    )

    # Merge cost, duration and resources
    final_df = cost_df.merge(duration_df, on="task", how="outer")
    final_df = final_df.merge(resources_df, on="task", how="outer")
    return final_df


def compare_omics_runs(run_ids, session, output_path):
    # Create the output directory if it doesn't exist
    if output_path:
        os.makedirs(output_path, exist_ok=True)
    else:
        output_path = os.getcwd()

    all_df = []
    for run_id in run_ids:
        single_df = single_run(run_id, session, output_path)
        all_df.append(single_df)

    final_df = all_df[0]
    for single_df in all_df[1:]:
        final_df = final_df.merge(single_df, on="task", how="outer")

    compare_file = f"{output_path}/compare_omics_runs.csv"
    final_df.to_csv(compare_file, index=False, float_format="%.3f")
    print(f"Comparison saved in: {compare_file}")


def main():
    parser = ArgumentParser()
    parser.add_argument("--region", type=str, help="AWS region to use", default="us-east-1")
    parser.add_argument(
        "--output-path",
        type=str,
        help="Output path for all files copied and generated during the analysis",
        default=None,
    )
    parser.add_argument("--run-ids", type=str, help="Omics run ids to compare (seprated by comma)", required=True)

    args = parser.parse_args()
    session = boto3.Session(region_name=args.region)
    run_ids = args.run_ids.split(",")

    compare_omics_runs(run_ids, session, args.output_path)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# noqa
"""
NOTE: This script is a copy of the original calculate_cost.py script from terra_pipeline.
"""

import argparse
import gzip
import json
import re
import sys
from datetime import timedelta
from io import BytesIO
from math import ceil
from urllib.request import urlopen

# pip install python-dateutil
import dateutil.parser

GCE_MACHINE_TYPES_URL = "http://cloudpricingcalculator.appspot.com/static/data/pricelist.json"
TOTAL_WORKFLOW_COST = 0

CUSTOM_MACHINE_CPU = "CP-COMPUTEENGINE-CUSTOM-VM-CORE"
CUSTOM_MACHINE_RAM = "CP-COMPUTEENGINE-CUSTOM-VM-RAM"
CUSTOM_MACHINE_EXTENDED_RAM = "CP-COMPUTEENGINE-CUSTOM-VM-EXTENDED-RAM"
CUSTOM_MACHINE_CPU_PREEMPTIBLE = "CP-COMPUTEENGINE-CUSTOM-VM-CORE-PREEMPTIBLE"
CUSTOM_MACHINE_RAM_PREEMPTIBLE = "CP-COMPUTEENGINE-CUSTOM-VM-RAM-PREEMPTIBLE"
CUSTOM_MACHINE_EXTENDED_RAM_PREEMPTIBLE = "CP-COMPUTEENGINE-CUSTOM-VM-EXTENDED-RAM-PREEMPTIBLE"
CUSTOM_MACHINE_TYPES = [
    CUSTOM_MACHINE_CPU,
    CUSTOM_MACHINE_RAM,
    CUSTOM_MACHINE_EXTENDED_RAM,
    CUSTOM_MACHINE_CPU_PREEMPTIBLE,
    CUSTOM_MACHINE_RAM_PREEMPTIBLE,
    CUSTOM_MACHINE_EXTENDED_RAM_PREEMPTIBLE,
]
GPU_NVIDIA_TESLA = "GPU_NVIDIA_TESLA"


# load the US pricing for both persistent disk and compute engine
def get_gce_pricing():  # noqa: C901
    response = urlopen(GCE_MACHINE_TYPES_URL)  # noqa: S310
    data = response.read()

    if response.info().get("Content-Encoding") == "gzip":
        buf = BytesIO(data)
        f = gzip.GzipFile(fileobj=buf)
        data = f.read()

    pricing = json.loads(data)

    data = {}
    for k, v in pricing.items():
        if k == "gcp_price_list":
            for k2, v2 in v.items():
                if k2.startswith("CP-COMPUTEENGINE-VMIMAGE"):
                    data[k2.replace("CP-COMPUTEENGINE-VMIMAGE-", "").lower()] = v2["us"]
                if k2 == "CP-COMPUTEENGINE-STORAGE-PD-SSD":
                    data[k2] = v2["us-central1"]
                if k2.startswith("CP-COMPUTEENGINE-LOCAL-SSD"):
                    data[k2] = v2["us-central1"]
                if k2.startswith("CP-COMPUTEENGINE-LOCAL-SSD-PREEMPTIBLE"):
                    data[k2] = v2["us-central1"]
                if k2 == "CP-COMPUTEENGINE-STORAGE-PD-CAPACITY":
                    data[k2] = v2["us-central1"]
                if k2 in CUSTOM_MACHINE_TYPES:
                    data[k2] = v2["us-central1"]
                if k2.startswith("GPU_NVIDIA_TESLA"):
                    data[k2] = v2["us-central1"]

    return data


def extract_machine_type(t):
    if "jes" in t and "machineType" in t["jes"]:
        full_machine = t["jes"]["machineType"]
        if full_machine.startswith("custom"):
            return "custom", full_machine
        else:
            return full_machine.split("/")[1], full_machine
    else:
        return "unknown", "unknown"


def extract_gpu_type(t):
    if "runtimeAttributes" in t and "gpuType" in t["runtimeAttributes"]:
        full_machine = t["runtimeAttributes"]["gpuType"]
        if full_machine.startswith("nvidia"):
            return full_machine
    else:
        return "unknown"


def get_disk_info(metadata):
    if "runtimeAttributes" in metadata and "disks" in metadata["runtimeAttributes"]:
        bootDiskSizeGb = 0.0  # noqa: N806
        if "bootDiskSizeGb" in metadata["runtimeAttributes"]:
            bootDiskSizeGb = float(metadata["runtimeAttributes"]["bootDiskSizeGb"])  # noqa: N806
        # Note - am lumping boot disk in with requested disk.  Assuming boot disk is same type as requested.
        # i.e. is it possible that boot disk is HDD when requested is SDD.
        (name, disk_size, disk_type) = metadata["runtimeAttributes"]["disks"].split()
        return {"size": float(disk_size) + bootDiskSizeGb, "type": "PERSISTENT_" + disk_type}
    else:
        # we can't tell disk size in this case so just return nothing
        return {"size": float(0), "type": "PERSISTENT_SSD"}


def was_preemptible_vm(metadata):
    if "runtimeAttributes" in metadata and "preemptible" in metadata["runtimeAttributes"]:
        pe_count = int(metadata["runtimeAttributes"]["preemptible"])
        attempt = int(metadata["attempt"])

        return attempt <= pe_count
    else:
        # we can't tell (older metadata) so conservatively return false
        return False


def used_cached_results(metadata):
    return "callCaching" in metadata and metadata["callCaching"]["hit"]


def calculate_runtime(call_info, ignore_preempted):
    # get start (start time of VM start) & end time (end time of 'ok') according to metadata

    # give a runtime of 0 for preempted jobs so they have no cost associated with them
    if was_preempted(call_info) and ignore_preempted:
        return 0

    # Accumulate the time spent in the following, which are the ones that we pay for
    pattern = re.compile("Pulling")
    total_amount = timedelta()
    if "executionEvents" in call_info:
        jobs_to_sum = [
            "UserAction",
            "Delocalization",
            "ContainerSetup",
            "Localization",
            "Complete in GCE / Cromwell Poll Interval",
            "^Pulling*",
        ]
        for x in call_info["executionEvents"]:
            if "description" in x:
                y = x["description"]
                if y in jobs_to_sum or pattern.match(y) is not None:
                    start = dateutil.parser.parse(x["startTime"])
                    end = dateutil.parser.parse(x["endTime"])
                    total_amount += end - start
        elapsed = total_amount
    # if we are preempted or if cromwell used previously cached results, we don't even get a start time from JES.
    # if cromwell was restarted, the start time from JES might not have been written to the metadata.
    # in either case, use the Cromwell start time which is earlier but not wrong.
    else:
        start = dateutil.parser.parse(call_info["start"])
        end = dateutil.parser.parse(call_info["end"])
        elapsed = end - start

    # The minimum runtime is 1 minute, after that it's by the second.
    # so if the task ran for 30 seconds, you pay for 1 minute.  If it ran for 1:01 minute, you pay for 1:01 minute
    seconds = elapsed.days * 24 * 60 * 60 + elapsed.seconds
    run_seconds = max(60.0, seconds)
    return run_seconds


def was_preempted(call_info):
    # We treat Preempted and RetryableFailure the same.  The latter is a general case of the former
    return call_info["executionStatus"] in ["Preempted", "RetryableFailure"]


def calculate_cost(metadata, ignore_preempted, only_total_cost, *, print_header, output_file=None):  # noqa: C901, PLR0912, PLR0915
    pipe_cost = 0
    # set up pricing information
    pricing = get_gce_pricing()
    ssd_cost_per_gb_per_month = float(pricing["CP-COMPUTEENGINE-STORAGE-PD-SSD"])
    ssd_cost_per_gb_hour = ssd_cost_per_gb_per_month / (24 * 365 / 12)

    local_ssd_cost_per_gb_per_month = float(pricing["CP-COMPUTEENGINE-LOCAL-SSD"])  # noqa: F841
    local_ssd_cost_per_gb_hour = ssd_cost_per_gb_per_month / (24 * 365 / 12)

    pe_local_ssd_cost_per_gb_per_month = float(pricing["CP-COMPUTEENGINE-LOCAL-SSD-PREEMPTIBLE"])  # noqa: F841
    pe_local_ssd_cost_per_gb_hour = ssd_cost_per_gb_per_month / (24 * 365 / 12)

    hdd_cost_per_gb_per_month = float(pricing["CP-COMPUTEENGINE-STORAGE-PD-CAPACITY"])
    hdd_cost_per_gb_hour = hdd_cost_per_gb_per_month / (24 * 365 / 12)

    disk_costs = {
        "PERSISTENT_SSD": ssd_cost_per_gb_hour,
        "PERSISTENT_HDD": hdd_cost_per_gb_hour,
        "PERSISTENT_LOCAL": local_ssd_cost_per_gb_hour,
        "PE_PERSISTENT_LOCAL": pe_local_ssd_cost_per_gb_hour,
    }

    if print_header and not only_total_cost:
        # print out a header
        header = ",".join(
            [
                "task_name",
                "status",
                "machine_type",
                "cpus",
                "machine_cpus",
                "mem_gbs",
                "total_hours",
                "cpu_cost_per_hour",
                "cpu_cost",
                "mem_cost_per_hour",
                "mem_cost",
                "pe_total_hours",
                "pe_cpu_cost_per_hour",
                "pe_cpu_cost",
                "pe_mem_cost_per_hour",
                "pe_mem_cost",
                "failed_pe_total_hours",
                "failed_pe_cpu_cost",
                "failed_pe_mem_cost",
                "disk_type",
                "disk_size",
                "disk_gb_hours",
                "disk_cost",
                "failed_pe_ssd_gb_hours",
                "failed_pe_ssd_cost",
                "gpu_cost",
                "pe_gpu_cost",
                "total_cost",
            ]
        )

        if output_file:
            with open(output_file, "a") as f:
                print(header, file=f)
        else:
            print(header)

    # iterate through the metadata file for each call
    for k, v in metadata["calls"].items():
        task_name = k

        total_hours = 0
        pe_total_hours = 0
        failed_pe_total_hours = 0
        cpus = 0
        mem_gbs = 0
        gpus = 0
        machine_type = "unknown"
        machine_name = "unknown"
        complete = True
        disk_info = get_disk_info({})
        gpu_type = "unknown"
        for call_info in v:
            # this is a subworkflow, recursively calculate cost on workflow metadata
            if "subWorkflowMetadata" in call_info:
                pipe_cost += calculate_cost(
                    call_info["subWorkflowMetadata"],
                    ignore_preempted,
                    only_total_cost,
                    print_header=False,
                    output_file=output_file,
                )
            # only process things that are not in flight
            elif call_info["executionStatus"] in ["Running", "NotStarted", "Starting"]:
                complete = False
            else:
                if call_info["executionStatus"] in ["Failed"]:
                    complete = False

                if machine_type == "unknown":
                    machine_type, machine_name = extract_machine_type(call_info)

                pe_vm = was_preemptible_vm(call_info)
                disk_info = get_disk_info(call_info)

                run_hours = calculate_runtime(call_info, ignore_preempted) / (60.0 * 60.0)

                # for preemptible VMs, separately tally successful tasks vs ones that were preempted
                if pe_vm:
                    if was_preempted(call_info):
                        # If Compute Engine terminates a preemptible instance less than 10 minutes after it is created,
                        # you are not billed for the use of that virtual machine instance
                        if run_hours < (10.0 / 60.0):
                            run_hours = 0
                        failed_pe_total_hours += run_hours
                    else:
                        pe_total_hours += run_hours
                else:
                    total_hours += run_hours
        # Runtime parameters are the same across all calls; just pull the info from the first one
        if "runtimeAttributes" in v[0]:
            if "cpu" in v[0]["runtimeAttributes"]:
                cpus += int(v[0]["runtimeAttributes"]["cpu"])
            if "memory" in v[0]["runtimeAttributes"]:
                mem_str = v[0]["runtimeAttributes"]["memory"]
                mem_gbs += float(mem_str[: mem_str.index(" ")])
            if "gpuCount" in v[0]["runtimeAttributes"]:
                gpus += int(v[0]["runtimeAttributes"]["gpuCount"])
            if "gpuType" in v[0]["runtimeAttributes"]:
                gpu_type = extract_gpu_type(v[0])

        if complete:
            status = "complete"
        else:
            status = "incomplete"

        machine_cpus = 0
        if machine_type == "custom":
            machine_cpus = int(machine_name.split("-")[1])

        if machine_type != "custom" and machine_type not in pricing:
            if "n2" in machine_type:
                machine_type = machine_type.replace("n2", "n1")
                if machine_type not in pricing:
                    machine_type = "unknown"
            else:
                machine_type = "unknown"

        if machine_type == "unknown":
            cpu_cost_per_hour = 0
            pe_cpu_cost_per_hour = 0
            mem_cost_per_hour = 0
            pe_mem_cost_per_hour = 0
        elif machine_type == "custom":
            cpu_cost_per_hour = pricing[CUSTOM_MACHINE_CPU] * cpus
            pe_cpu_cost_per_hour = pricing[CUSTOM_MACHINE_CPU_PREEMPTIBLE] * cpus
            mem_cost_per_hour = pricing[CUSTOM_MACHINE_RAM] * mem_gbs
            pe_mem_cost_per_hour = pricing[CUSTOM_MACHINE_RAM_PREEMPTIBLE] * mem_gbs
        else:
            cpu_cost_per_hour = pricing[machine_type]
            pe_cpu_cost_per_hour = pricing[machine_type + "-preemptible"]
            mem_cost_per_hour = 0
            pe_mem_cost_per_hour = 0

        gpu_cost_per_hour = 0
        pe_gpu_cost_per_hour = 0
        if gpu_type.startswith("nvidia-tesla-"):
            nvidia_type = gpu_type.split("-")[2]
            full_gpu_type = GPU_NVIDIA_TESLA + "_" + nvidia_type.upper()
            pe_full_gpu_type = GPU_NVIDIA_TESLA + "_" + nvidia_type.upper() + "-PREEMPTIBLE"
            if full_gpu_type in pricing:
                gpu_cost_per_hour = pricing[full_gpu_type] * gpus
                pe_gpu_cost_per_hour = pricing[pe_full_gpu_type] * gpus

        cpu_cost = total_hours * cpu_cost_per_hour
        failed_pe_cpu_cost = failed_pe_total_hours * pe_cpu_cost_per_hour
        pe_cpu_cost = pe_total_hours * pe_cpu_cost_per_hour

        gpu_cost = total_hours * gpu_cost_per_hour
        failed_pe_gpu_cost = failed_pe_total_hours * pe_gpu_cost_per_hour
        pe_gpu_cost = pe_total_hours * pe_gpu_cost_per_hour

        #
        # NOTE -- local ssds have a different price when used in preemptible VMs.
        # However, to implement this all the disk calculations need to be moved from the task level (where it is now)
        # to the call level since each call could be preemptible or not
        # Then we can decide to use PERSISTENT_LOCAL or PE_PERSISTENT_LOCAL
        #
        disk_cost_per_gb_hour = disk_costs[disk_info["type"]]

        if disk_info["type"].endswith("PERSISTENT_LOCAL"):
            local_disk_fix_size = 375
            disk_size = ceil(disk_info["size"] / local_disk_fix_size) * local_disk_fix_size
        else:
            disk_size = disk_info["size"]

        disk_gb_hours = disk_size * (total_hours + pe_total_hours)
        disk_cost = disk_gb_hours * disk_cost_per_gb_hour

        failed_pe_disk_gb_hours = disk_size * failed_pe_total_hours
        failed_pe_disk_cost = failed_pe_disk_gb_hours * disk_cost_per_gb_hour

        mem_cost = total_hours * mem_cost_per_hour
        pe_mem_cost = pe_total_hours * pe_mem_cost_per_hour
        failed_pe_mem_cost = failed_pe_total_hours * pe_mem_cost_per_hour

        total_cost = (
            cpu_cost
            + pe_cpu_cost
            + failed_pe_cpu_cost
            + gpu_cost
            + pe_gpu_cost
            + failed_pe_gpu_cost
            + disk_cost
            + failed_pe_disk_cost
            + mem_cost
            + pe_mem_cost
            + failed_pe_mem_cost
        )

        # accumalate total workflow cost
        global TOTAL_WORKFLOW_COST  # noqa: PLW0603
        TOTAL_WORKFLOW_COST += total_cost
        pipe_cost += total_cost

        if not only_total_cost:
            out = (
                task_name,
                status,
                machine_type,
                cpus,
                machine_cpus,
                mem_gbs,
                total_hours,
                cpu_cost_per_hour,
                cpu_cost,
                mem_cost_per_hour,
                mem_cost,
                pe_total_hours,
                pe_cpu_cost_per_hour,
                pe_cpu_cost,
                pe_mem_cost_per_hour,
                pe_mem_cost,
                failed_pe_total_hours,
                failed_pe_cpu_cost,
                failed_pe_mem_cost,
                disk_info["type"],
                disk_size,
                disk_gb_hours,
                disk_cost,
                failed_pe_disk_gb_hours,
                failed_pe_disk_cost,
                gpu_cost,
                pe_gpu_cost,
                total_cost,
            )
            if output_file:
                with open(output_file, "a") as f:
                    print(",".join(map(str, out)), file=f)
            else:
                print(",".join(map(str, out)))

    return pipe_cost


def compare(old, new):
    """Fail when NEW total exceeds OLD total by > 5%."""

    def total(cost_file):
        with open(cost_file) as input:  # noqa: A001
            lines = input.readlines()
        for line in lines:
            fields = line.split()
            if len(fields) == 3 and fields[0] == "Total" and fields[1] == "Cost:":  # noqa: PLR2004
                return int(float(fields[2]) * 10000) / 10000.0
        return None

    old_cost = total(old)
    new_cost = total(new)
    if old_cost and new_cost:
        more = new_cost - old_cost
        percent = more * 100 / old_cost
        if more > 0.0:
            print(f"Cost has increased by ${more} ({percent}%): from ${old_cost} to ${new_cost}")
            sys.exit(0)
        else:
            if more < 0.0:
                down = percent * -1
                print(f"Cost has decreased by ${more} ({down}%): from ${old_cost} to ${new_cost}")
            else:
                print(f"Cost is the same: ${new_cost}")
            print("Everything is awesome!")
            sys.exit(0)
    sys.exit(f"One or both of the calculated costs is 0!  WTF?: old ({old_cost}) new ({new_cost})")


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--ignore_preempted", dest="ignore_preempted", action="store_true", help="ignore preempted tasks"
    )
    parser.add_argument(
        "--only_total",
        dest="only_total_cost",
        action="store_true",
        help="print total cost of the workflow instead of the tsv per task costs",
    )
    either = parser.add_mutually_exclusive_group(required=True)
    either.add_argument("-m", "--metadata", dest="metadata", help="metadata file to calculate cost on")
    either.add_argument("--compare", nargs=2, help="compare old to new cost output")

    args = parser.parse_args()

    if args.metadata:
        with open(args.metadata) as data_file:
            metadata = json.load(data_file)
        calculate_cost(metadata, args.ignore_preempted, args.only_total_cost, print_header=True)
        if args.only_total_cost:
            print("Total Cost: " + str(TOTAL_WORKFLOW_COST))
    else:
        old, new = args.compare
        compare(old, new)


if __name__ == "__main__":
    main()



import os
import subprocess
from enum import Enum

import pandas as pd

RUN_ID_PLACEHOLDER = "<RUN_ID>"
PLOTS_DIR = f"omics_{RUN_ID_PLACEHOLDER}_plots"


class Columns(Enum):
    ESTIMATED_USD_COLUMN = "estimatedUSD"
    TYPE_COLUMN = "type"
    NAME_COLUMN = "name"
    OMICS_INSTANCE_TYPE_RESERVED = "omicsInstanceTypeReserved"
    CPU_REQUESTED = "cpusRequested"
    MEMORY_REQUESTED_GIB = "memoryRequestedGiB"
    GPUS_REQUESTED = "gpusRequested"


class RunCost:
    def __init__(self, run_id, output_dir=None, output_prefix=""):
        self.run_id = run_id
        self.output_dir = output_dir
        if self.output_dir:
            os.makedirs(output_dir, exist_ok=True)
        self.output_prefix = output_prefix
        self.cost_csv = None
        self.cost_df = None

    def calculate_run_cost(self) -> None:
        cost_output = f"{self.output_prefix}omics_{self.run_id}.cost.csv"
        plots_dir = PLOTS_DIR.replace(RUN_ID_PLACEHOLDER, self.run_id)

        if self.output_dir:
            cost_output = f"{self.output_dir}/{cost_output}"
            plots_dir = f"{self.output_dir}/{plots_dir}"

        self.run_analyzer(self.run_id, cost_output, plots_dir)
        print(f"Run cost calculated and saved to {cost_output}")
        self.cost_csv = cost_output
        self.cost_df = pd.read_csv(self.cost_csv)

    def get_total_cost(self) -> float:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[Columns.ESTIMATED_USD_COLUMN.value].sum()

    def get_storage_cost(self) -> float:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[self.cost_df[Columns.TYPE_COLUMN.value] == "run"][
            Columns.ESTIMATED_USD_COLUMN.value
        ].to_numpy()[0]

    def get_tasks_cost(self) -> pd.DataFrame:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[self.cost_df[Columns.TYPE_COLUMN.value] == "task"][
            [Columns.NAME_COLUMN.value, Columns.ESTIMATED_USD_COLUMN.value, Columns.OMICS_INSTANCE_TYPE_RESERVED.value]
        ]

    def get_tasks_resources(self) -> pd.DataFrame:
        if self.cost_df is None:
            self.calculate_run_cost()

        return self.cost_df[self.cost_df[Columns.TYPE_COLUMN.value] == "task"][
            [
                Columns.NAME_COLUMN.value,
                Columns.OMICS_INSTANCE_TYPE_RESERVED.value,
                Columns.CPU_REQUESTED.value,
                Columns.MEMORY_REQUESTED_GIB.value,
                Columns.GPUS_REQUESTED.value,
            ]
        ]

    @staticmethod
    def run_analyzer(run_id, run_csv_name, plots_dir):
        analyzer_command = ["python", "-m", "omics.cli.run_analyzer", run_id, "-o", run_csv_name, "-P", plots_dir]

        subprocess.run(
            analyzer_command,
            check=True,
        )



import os
from argparse import ArgumentParser

import boto3

from ugbio_omics.get_run_info import get_run_info

FAILED_STATUS = "FAILED"
OMICS_LOG_GROUP = "/aws/omics/WorkflowLog"


def get_log_for_task(run_id, task_id=None, session=None, output_path=None, output_prefix="", *, failed: bool = True):
    # Get omics client to retrieve the run and tasks information
    if session:
        omics_client = session.client("omics")
    else:
        omics_client = boto3.client("omics")
    run = get_run_info(run_id, client=omics_client)

    # If no given a specific task_id, get the logs for all tasks or failed tasks
    if not task_id:
        task_ids = [task["taskId"] for task in run["tasks"] if (not failed or task["status"] == FAILED_STATUS)]
        task_names = {
            task["taskId"]: task["name"] for task in run["tasks"] if (not failed or task["status"] == FAILED_STATUS)
        }

    else:
        task_ids = [task_id]
        task_names = {task["taskId"]: task["name"] for task in run["tasks"] if task["taskId"] == task_id}

    # Get the logs for each task
    for task_id_val in task_ids:
        task_name = task_names[task_id_val]
        print("------------------------------------------")
        print(f"Getting log for task {task_name} (taskId: {task_id_val})")
        log_stream_name = f"run/{run_id}/task/{task_id_val}"

        output = f"{output_prefix}run_{run_id}_task_{task_id_val}_{task_name}.log"
        fetch_save_log(log_stream_name, output, output_path, session)

    # in case that the run failed but there're no failed tasks the run's engine log should include the error
    if run["status"] == FAILED_STATUS and FAILED_STATUS not in [task["status"] for task in run["tasks"]]:
        print(f"Run status is {FAILED_STATUS} but no failed tasks. Getting run's engine log")
        log_stream_name = f"run/{run_id}/engine"
        output = f"{output_prefix}run_{run_id}_engine.log"
        fetch_save_log(log_stream_name, output, output_path, session)


def fetch_save_log(log_stream_name, output, output_path, session=None):
    # Get logs client
    if session:
        client = session.client("logs")
    else:
        client = boto3.client("logs")

    print(f"Getting log events for log group '{OMICS_LOG_GROUP}' and log stream '{log_stream_name}'")

    # get first page of log events
    response = client.get_log_events(logGroupName=OMICS_LOG_GROUP, logStreamName=log_stream_name, startFromHead=True)

    # check if events is not empty
    if not response.get("events"):
        print(f"No events found for log group '{OMICS_LOG_GROUP}' and log stream '{log_stream_name}'")
        return

    if output_path:
        os.makedirs(output_path, exist_ok=True)
        output = f"{output_path}/{output}"

    # write event message to output file
    with open(output, "w") as file:
        for event in response["events"]:
            file.write(f"{event['message']}\n")

        # get next page of log events
        while len(response.get("events")) > 0 and response.get("nextForwardToken"):
            response = client.get_log_events(
                logGroupName=OMICS_LOG_GROUP,
                logStreamName=log_stream_name,
                nextToken=response.get("nextForwardToken"),
                startFromHead=True,
            )
            for event in response["events"]:
                file.write(f"{event['message']}\n")

    print(f"Log file saved to: {output}")
    return output


def main():
    parser = ArgumentParser()
    parser.add_argument("--region", type=str, help="AWS region to use", default="us-east-1")
    parser.add_argument("--run-id", type=str, help="HealthOmics workflow run-id to analyze")
    parser.add_argument(
        "--task-id",
        type=str,
        help="HealthOmics workflow task-id to analyze. Leave empty to get the logs for all tasks",
        default=None,
    )
    parser.add_argument(
        "--failed", dest="failed", action="store_true", help="Set to true to get logs for failed tasks only"
    )
    parser.add_argument("--no-failed", dest="failed", action="store_false")
    parser.set_defaults(failed=False)
    parser.add_argument("--output", type=str, help="Output dir to save log events", default=None)
    parser.add_argument("--output-prefix", type=str, help="File name prefix for the output", required=False, default="")

    args = parser.parse_args()
    session = boto3.Session(region_name=args.region)

    get_log_for_task(
        args.run_id,
        args.task_id,
        session=session,
        output_path=args.output,
        output_prefix=args.output_prefix,
        failed=args.failed,
    )


if __name__ == "__main__":
    main()
